package ru.dbotthepony.mc.prng;

import com.mojang.datafixers.util.Pair;
import net.minecraft.core.HolderLookup;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.NbtOps;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.util.RandomSource;
import net.minecraft.world.RandomSequence;
import net.minecraft.world.RandomSequences;
import net.minecraft.world.level.saveddata.SavedData;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.jetbrains.annotations.NotNull;

import java.util.HashMap;
import java.util.Optional;
import java.util.function.BiConsumer;

public final class BetterRandomSequences extends RandomSequences {
	private static final Logger LOGGER = LogManager.getLogger();

	public BetterRandomSequences(long seed) {
		super(seed);
	}

	public BetterRandomSequences(long seed, CompoundTag tag, HolderLookup.Provider provider) {
		super(seed);

		includeWorldSeed = tag.getBoolean("includeWorldSeed");
		includeSequenceId = tag.getBoolean("includeSequenceId");
		salt = tag.getInt("salt");

		var compound = tag.getCompound("sequences");

		for (var key : compound.getAllKeys()) {
			try {
				var location = ResourceLocation.parse(key);
				var rng = GJRAND64RandomSource.CODEC.decode(NbtOps.INSTANCE, compound.get(key)).map(Pair::getFirst).getOrThrow();
				betterSequences.put(location, new BetterRandomSequence(rng));
			} catch (RuntimeException err) {
				LOGGER.error("Unable to deserialize random sequence at {}", key, err);
			}
		}
	}

	private final HashMap<ResourceLocation, BetterRandomSequence> betterSequences = new HashMap<>();

	@Override
	public @NotNull CompoundTag save(@NotNull CompoundTag tag, HolderLookup.@NotNull Provider registries) {
		// super.save(tag, registries);
		tag.putBoolean("includeWorldSeed", includeWorldSeed);
		tag.putBoolean("includeSequenceId", includeSequenceId);
		tag.putInt("salt", salt);

		var compound = new CompoundTag();
		tag.put("sequences", compound);

		for (var entry : betterSequences.entrySet()) {
			compound.put(entry.getKey().toString(), GJRAND64RandomSource.CODEC.encodeStart(NbtOps.INSTANCE, entry.getValue().actualRandom).getOrThrow());
		}

		return tag;
	}

	@Override
	public @NotNull RandomSource get(@NotNull ResourceLocation location) {
		return new DirtyMarkingRandomSource(betterSequences.computeIfAbsent(location, this::createSequence).actualRandom);
	}

	private BetterRandomSequence createSequence(ResourceLocation location) {
		return this.createSequence(location, this.salt, this.includeWorldSeed, this.includeSequenceId);
	}

	private BetterRandomSequence createSequence(ResourceLocation location, int salt, boolean includeWorldSeed, boolean includeSequenceId) {
		long i = (includeWorldSeed ? this.worldSeed : 0L) ^ (long)salt;
		return new BetterRandomSequence(i, includeSequenceId ? Optional.of(location) : Optional.empty());
	}

	@Override
	public void forAllSequences(@NotNull BiConsumer<ResourceLocation, RandomSequence> action) {
		betterSequences.forEach(action);
	}

	@Override
	public int clear() {
		int i = betterSequences.size();
		betterSequences.clear();
		return i;
	}

	@Override
	public void reset(@NotNull ResourceLocation sequence) {
		betterSequences.put(sequence, createSequence(sequence));
	}

	@Override
	public void reset(@NotNull ResourceLocation sequence, int seed, boolean includeWorldSeed, boolean includeSequenceId) {
		betterSequences.put(sequence, createSequence(sequence, seed, includeWorldSeed, includeSequenceId));
	}

	public static SavedData.Factory<BetterRandomSequences> betterFactory(final long seed) {
		return new SavedData.Factory<>(
			() -> new BetterRandomSequences(seed),
			(tag, provider) -> new BetterRandomSequences(seed, tag, provider),
			null
		);
	}
}
