package ru.dbotthepony.mc.prng.mixin;

import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.util.RandomSource;
import net.minecraft.world.RandomSequences;
import net.minecraft.world.level.Level;
import org.jetbrains.annotations.Nullable;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Overwrite;
import ru.dbotthepony.mc.prng.BetterRandom;
import ru.dbotthepony.mc.prng.BetterRandomSequences;

import java.util.Objects;

@Mixin(ServerLevel.class)
public abstract class ServerLevelMixin {
	@Nullable
	private BetterRandomSequences betterRandomSequences;

	private ServerLevel self$brng() {
		return (ServerLevel) (Object) this;
	}

	@Overwrite(remap = false)
	public RandomSource getRandomSequence(ResourceLocation location) {
		return getRandomSequences().get(location);
	}

	@Overwrite(remap = false)
	public RandomSequences getRandomSequences() {
		if (this.betterRandomSequences == null) {
			ServerLevel overworld = Objects.requireNonNull(
				self$brng()
				.getServer()
				.getLevel(Level.OVERWORLD), "MinecraftServer lacks an Overworld Level, cannot continue. (exception thrown by BetterRandom)");

			this.betterRandomSequences = overworld
				.getDataStorage()
				.computeIfAbsent(BetterRandomSequences.betterFactory(overworld.getSeed()), BetterRandom.SAVEDATA_LOCATION);
		}

		return this.betterRandomSequences;
	}
}
